/*****************************************************************************/
/*									     */
/*									     */
/*	Xpack version 1.00 -- module Xaw-main.c				     */
/*									     */
/*	main function for the Athena Widget interface.			     */
/*	Written by Michael Bischoff (mbi@mo.math.nat.tu-bs.de)		     */
/*	November-1994							     */
/*	see COPYRIGHT.xpack for Copyright details			     */
/*									     */
/*									     */
/*****************************************************************************/
#include "X-sok.h"
#include "Tableau.h"
#include "version.h"

Widget toplevel;
const char *xpackdir;
struct game game;

static XtAppContext app_con;
static void (*execfunc)(void) = NULL;
static Widget messagebox, container, desktop;
static Widget dialog, popup, paned;
static Window mainwindow;


void show_message(const char *str, ...) {
    if (gamegraphic) {
	static char last_message[256];
	Arg Args;
	va_list args;
	va_start(args, str);
	
	if (!str) {
	    memset(last_message, ' ', sizeof(last_message)-1);
	    last_message[sizeof(last_message)-1] = '\0';
	} else
	    vsprintf(last_message, str, args);
	
	XtSetArg(Args, XtNlabel, last_message);
	XtSetValues(messagebox, &Args, 1);
    }
}

void SetTitle(void) {
    if (XtWindow(toplevel)) {
	static char windowname[48];
	sprintf(windowname, "%s", P.description);
	XStoreName(dpy, XtWindow(toplevel), windowname);
    }
}

void cmd_LeavePack(void) {
    XtDestroyApplicationContext(app_con);
    exit(0);
}

static void perform_command(Widget widget, XtPointer client_data, XtPointer call_data) {
    (*(void (*)(void))client_data)();	/* any questions? */
}

static void popup_confirm(const char *prompt) {
    Arg args[2];
    Position x, y;
    Dimension xx, yy, wx, wy;

    XtSetArg(args[0], XtNlabel, prompt);
    XtSetValues(dialog, args, 1);
    XtVaGetValues(dialog,   XtNwidth, &xx, XtNheight, &yy, NULL);
    XtVaGetValues(toplevel, XtNwidth, &wx, XtNheight, &wy, NULL);
    x = (wx - xx)/2;
    y = (wy - yy)/2;

    XtTranslateCoords(toplevel, x, y, &x, &y);
    XtSetArg(args[0], XtNx, x);
    XtSetArg(args[1], XtNy, y);
    XtSetValues(popup, args, 2);
    XtPopup(popup, XtGrabNone);
}

void cmd_Confirm(void) {
    if (execfunc) {
	void (*execfunc2)(void) = execfunc;	/* erase it first! */
	execfunc = NULL;
	XtPopdown(popup);
	(*execfunc2)();	/* finally execute the desired function */
    }
}
void cmd_Cancel(void) {
    if (execfunc) {
	execfunc = NULL;
	XtPopdown(popup);
    }
}

/* type converter functions: */
static void mXtAP_Cancel (Widget w, XEvent *xev, String *params, Cardinal *num) { cmd_Cancel();  }
static void mXtAP_Confirm(Widget w, XEvent *xev, String *params, Cardinal *num) { cmd_Confirm(); }
static void Cancel(Widget widget, XtPointer client_data, XtPointer call_data)	{ cmd_Cancel();  }
static void Ok(Widget widget, XtPointer client_data, XtPointer call_data)	{ cmd_Confirm(); }

void request_confirm(void (*dofunc)(void), const char *prompt) {
    if (execfunc)
	return;		/* request pending => deny another one */
    if (game.finished || !game.n_moves) {
	(*dofunc)();	/* perform action without confirmation */
	return;
    }
    execfunc = dofunc;
    popup_confirm(prompt);
}

static String fallback_resources[] = { 
    "*beNiceToColormap:			false",
    "*shapeStyle:			Rectangle",
    "*topShadowContrast:		20",
    "*bottomShadowContrast:		40",
    "*Scrollbar*background:		Grey70",
    "*Background:			grey85",
    "*Foreground:			black",
    "*resizeToPreferred:		True",
    "*input:				True",
    "*showGrip:				off",
    "*shadowWidth:                   	2",
    "*messages.justify:			Left",
    "*upperbox.orientation:		XtorientHorizontal",
    "*lowerbox.orientation:		XtorientHorizontal",
    "*Tableau.backingStore:		WhenMapped",
    "*Tableau.keyboardFile:		keys",
    "*Tableau.messageFile:		messages",
    "*Tableau.background:		black",
    "*Tableau.rules:		        ",
    "*Viewport.allowHoriz:		False",		/* scrollbar disable! */
    "*Viewport.allowVert:		False",		/* code is buggy! */
    "*Viewport.forceBars:		False",
    "*Viewport.useBottom:		True",
    "*Viewport.useRight:		True",
    "*Label.shadowWidth:		0",
    "*Label.BorderWidth:		2",
    "*Dialog*Translations: #override \n<Key>y: Ok()\n<Key>n: Cancel()\n",
    "XPack*title:			XPack",
    "XPack.prompt.allowShellResize:	True",
    "XPack.prompt.saveUnder:		True",
    "*Dialog*resizable:			True",
    NULL,
};

static XrmOptionDescRec options[] = {
    /* tableau resources */
    { "-rules",		"*Tableau.rules",	XrmoptionSepArg, NULL },
    { "-level",		"*Tableau.level",	XrmoptionSepArg, NULL },
    { "-factor",	"*Tableau.factor",	XrmoptionSepArg, NULL },
    { "-diff",		"*Tableau.diff",	XrmoptionSepArg, NULL },
    { "-width",		"*Tableau.width",	XrmoptionSepArg, NULL },
    { "-height",	"*Tableau.height",	XrmoptionSepArg, NULL },
};

static XtActionsRec moreActions[] = {
    { "Cancel",		mXtAP_Cancel },
    { "Ok",		mXtAP_Confirm }
};

static void process_extra_args(int argc, char *argv[]) {
    /* check extra args */
    if (argc >= 2) {
	fprintf(stderr, "xpack: invalid argument: %s\n", argv[1]);
	fprintf(stderr, "usage: xpack [options]\n"
		"options are all standard X11 toolkit options and\n"
                "-rules (ruleset)     to initially use specified rules\n"
                "-level (levelnr)     to set the startlevel\n"
		);
	exit(EXIT_FAILURE);
    }
}

static void selecttile(Widget w, XtPointer number, XtPointer garbage) {
    const char *s = XtName(w);
    int i;
    if (state) {
	show_message("There is an active tile");
	return;
    }
    for (i = 1; i <= P.totaltiles; ++i)
	if (P.position[i].x < 0 && !strcmp(s, P.t[P.position[i].tilenr].name)) {
	    activate_free(i);
	    popup_active(0, 0, (DX+1) / 2, (DY+1) / 2);
	    state = 2;
	    show_message("OK");
	    return;
	}
    show_message("There is no free tile %s", s);
}

int main(int argc, char *argv[]) {
    Widget buttonpanel;
    Widget gamebutton, gamemenu, rulesbutton, rulesmenu;
    int i;
    struct button {
	const char *name; void (*func)(void);
    } *bp;
    static struct button buttons[] = {
	{ "Place Largest",	cmd_PlaceLargest },
	{ "Save",	  	cmd_SaveGame },
	{ "Load",		cmd_LoadGame }
    }, mbuttons[] = {
	{ "Restart",		cmd_RestartGame },
	{ "Try all charmap",	cmd_TryAll1 },
	{ "Try all intersect",	cmd_TryAll2 },
	{ "Try all bitmap",	cmd_TryAll3 },
	{ "Try all intbit",	cmd_TryAll4 },
	{ "Update if same",	cmd_SetImp0 },
	{ "Update if better",	cmd_SetImp1 },
	{ "Heuristik",		cmd_Heuristik1 },
	{ "Describe",		cmd_Describe },
	{ "LaTeX",		cmd_write_LaTeX },
	{ "PostScript",		cmd_write_PostScript },
	{ "Quit",		rq_LeavePack }
    };

    generate_rectangle_problem(1, 1, 0, 0, 0);
    make_simple_solution();

    /* use the command line arguments concerning the widgets */
    switch_uid(1);
    toplevel = XtAppInitialize(&app_con, "XPack", options, XtNumber(options),
			       &argc, argv, fallback_resources, NULL, 0);
    switch_uid(0);
    process_extra_args(argc, argv);
    XtAppAddActions(app_con, moreActions, XtNumber(moreActions));

    /* basic elements */
    paned       = XtCreateManagedWidget("paned",	panedWidgetClass,    toplevel,	  NULL, 0);
    buttonpanel = XtCreateManagedWidget("buttonpanel",	boxWidgetClass,     paned,	  NULL, 0);
    messagebox = XtCreateManagedWidget("messages", labelWidgetClass, paned, NULL, 0);
    show_message("Welcome to xpack version %s!", VERSION);
    graphics_control(Disable);

    container   = XtCreateManagedWidget("container",	viewportWidgetClass, paned,	  NULL, 0);
    desktop     = XtCreateManagedWidget("desktop",	tableauWidgetClass,  container,	  NULL, 0);
    /* XtAddCallback(container, XtNreportCallback, reportfunc, NULL); */

    /* create the button panel and its menus */
    gamebutton  = XtCreateManagedWidget("Algorithm", menuButtonWidgetClass, buttonpanel, NULL, 0);
    gamemenu    = XtCreatePopupShell("gamemenu", simpleMenuWidgetClass, gamebutton, NULL, 0);
    for (bp = mbuttons, i = 0; i < XtNumber(mbuttons); ++i) {
	Widget w;
	w = XtCreateManagedWidget(bp->name, smeBSBObjectClass, gamemenu, NULL, 0);
	if (bp->func) XtAddCallback(w, XtNcallback, perform_command, bp->func);
	++bp;
    }
    XtVaSetValues(gamebutton, XtNmenuName, "gamemenu", NULL);

    /* create Rules button just right of the Game button */
    rulesbutton = XtCreateManagedWidget("Select", menuButtonWidgetClass, buttonpanel, NULL, 0);
    rulesmenu   = XtCreatePopupShell("rulesmenu", simpleMenuWidgetClass, rulesbutton, NULL, 0);
    XtVaSetValues(rulesbutton, XtNmenuName, "rulesmenu", NULL);

    /* rest of the buttons */
    for (bp = buttons, i = 0; i < XtNumber(buttons); ++i) {
	Widget w;
	w = XtCreateManagedWidget(bp->name, commandWidgetClass, buttonpanel, NULL, 0);
	if (bp->func) XtAddCallback(w, XtNcallback, perform_command, bp->func);
	++bp;
    }

    /* OK. Now do the pop-up shells */
    popup = XtCreatePopupShell("prompt", transientShellWidgetClass, toplevel, NULL, 0);
    dialog = XtCreateManagedWidget("dialog", dialogWidgetClass, popup, NULL, 0);
    XawDialogAddButton(dialog, "ok",     Ok,     (XtPointer)dialog);
    XawDialogAddButton(dialog, "cancel", Cancel, (XtPointer)dialog);

    graphic.width = graphic.height = 0;
    graphic.autolayout = 1;
    XtRealizeWidget(toplevel);
    XSync(dpy, 0);
    mainwindow = XtWindow(toplevel);
    XSetIconName(dpy, mainwindow, "xpack");
    SetTitle();
    table = XtWindow(desktop);
    {   int i;
	for (i = 1; i <= P.maxtile; ++i) {
	    static char s[4];
	    Widget w;
	    sprintf(s, "%s", P.t[i].name);
	    w = XtCreateManagedWidget(s, smeBSBObjectClass, rulesmenu, NULL, 0);
	    XtAddCallback(w, XtNcallback, selecttile, NULL);
	}
    }

    graphics_control(Enable);
    XtRealizeWidget(popup);
    XtAppMainLoop(app_con);	/* does not return */
    return 0;			/* keep compiler happy */
}

void Force_Resize(XSize_t w, XSize_t h) {
    Arg args[1];
    int hh;
    Dimension hhh = 0;
    XtSetArg(args[0], XtNheight, &hhh);
    XtGetValues(paned, args, 1);
    hh = hhh;		/* unsigned short => int */
    XtGetValues(container, args, 1);
    /* printf("rq %d, paned = %d, container = %d\n", h, hh, hhh); */
    h += hh - hhh;	/* difference between overall size and Viewport size */

    XResizeWindow(dpy, mainwindow, w, h);
}
